@@ -7,12 +7,13 @@ import types as types
 from functools import wraps
 import hashlib as hashlib
 
+# support for numpy and pandas is optional for this module
+# At the moment both are listed as dependencies in setup.py to ensure
+# they are tested in github
+# April'20
 np = None
 pd = None
 
-# support for numpy and pandas is optional
-# for this module
-# April'20
 try:
     import numpy as np
 except:
@@ -158,7 +159,7 @@ def plain( inn, sorted = False ):
     if not pd is None and isinstance(inn,pd.DataFrame):
         plain(inn.columns)
         plain(inn.index)
-        plain(inn.as_matrix())
+        plain(inn.to_numpy())
         return
     # lists, tuples and everything which looks like it --> lists
     if not getattr(inn,"__iter__",None) is None: #isinstance(inn,list) or isinstance(inn,tuple):
@@ -174,52 +175,64 @@ def plain( inn, sorted = False ):
     # nothing we can do
     raise TypeError(fmt("Cannot handle type %s", type(inn)))
 
-def unqiueHash(*args, **kwargs):
+def uniqueHash(*args, **kwargs):
     """ Generates a hash key for any collection of python objects.
-        Typical use is for key'ing data vs a unique configuation
-        Hans Buehler 207
+        Typical use is for key'ing data vs a unique configuation.
+        
+        The function
+            1) uses the repr() function to feed objects to the hash algorithm.
+               that means is only distinguishes floats up to str conversion precision
+            2) keys of dictionaries, and sets are sorted to ensure equality of hashes
+               accross different memory setups of strings
+        
+        Hans Buehler 2017
     """    
     m = hashlib.md5()
+    def update(s):
+        m.update(repr(s).encode('utf-8'))
     def visit(inn):
         # basics
-        if isAtomic(inn) \
-            or isinstance(inn,datetime.time,datetime.date,datetime.datetime) \
-            or (False if np is None else isinstance(inn,np.ndarray)) \
-            or inn is None:
-            m.update(inn)
+        if inn is None:
             return
         # can't handle functions --> return None
         if isFunction(inn) or isinstance(inn,property):
             return None
-        # dictionaries
-        if isinstance(inn,dict):
-            inns = list(inn.keys)
-            inns.sort()            
-            for k in inn:
-                m.update(k)
-                visit(inn[k])
+        # basic elements
+        if isAtomic(inn):
+            update(inn)
+            return
+        # some standard types
+        if isinstance(inn,(datetime.time,datetime.date,datetime.datetime)):
+            update(inn)
+            return
+        # numpy
+        if not np is None and isinstance(inn,np.ndarray):
+            update(inn)
             return
         # pandas
         if not pd is None and isinstance(inn,pd.DataFrame):
-            m.update(inn.columns)
-            m.update(inn.index)
-            m.update(inn.as_matrix())
+            update(inn)
             return
+        # sets --> sorted list
+        if isinstance(inn,set):
+            inn = list(inn)
+            inn.sort()
         # lists, tuples and everything which looks like it --> lists
         if not getattr(inn,"__iter__",None) is None:
-            try:
-                for k in inn:
-                    visit(k)
-            except TypeError:
-                pass #
+            for k in inn:
+                visit(k)
             return
-        # handle objects as dictionaries, removing all functions
-        dct = getattr(inn,"__dict__",None)
-        if not dct is None:
-            visit(dct)
-            return
-        # nothing we can do
-        raise TypeError(fmt("Cannot handle type %s", type(inn)))
+        # dictionaries
+        if not isinstance(inn,dict):
+            inn = getattr(inn,"__dict__",None)
+            if inn is None: raise TypeError(fmt("Cannot handle type %s", type(inn)))
+        
+        inns = list(inn.keys())
+        inns.sort() # this ensures that dictionaries are always in the right order
+        for k in inn:
+            update(k)
+            visit(inn[k])
+        return
         
     visit(args)
     visit(kwargs)
@@ -263,29 +276,41 @@ def bind( F, **kwargs ):
 
 class Generic(object):
     """ An object which can be used as a generic store.
-        It allows 
-        
-        Constructions by keyword:
-        
+        Constructions by keyword:        
             g = Generic(a=1, b=2, c=3)
-            
-            
-        Access by key
-        
+                        
+        Access by key        
             a = g.a
             a = g['a']
             e = g.get('e',None)   # with default
             e = g('e',None)       # with default
-            
             del g.e
             
+        Such objects can be extended with functions e.g. you can set
+            def f(x):
+                return x*x
+            g.f = f
+            g.f(2) --> 4
+
+        This works as well with bound function, e.g. 
+            def F(self,a):
+                self.a = a            # set the value of the 'g' object
+            g.F = MethodType(F)
+            ...
+            g.F(1)
+            g.a --> 1
+            
+        
+        However, functions are _not_ copied when at construction or when merging one Generic
+        into another. This because the best u
+    
+            
     Hans Buehler, 2013
     """
     
     def __init__(self, *vargs, **kwargs):
         """ Initialize object; use keyword notation such as
-                Generic(a=1, b=2)
-                
+                Generic(a=1, b=2)                
             The 'vargs' argument allows passing on existing Generic, dict or generic objects;
             see merge() for further information
         """
@@ -295,34 +320,56 @@ class Generic(object):
     # make it behave like a dictionary    
     
     def __str__(self):
+        """ like dict """
         return self.__dict__.__str__()
     
     def __repr__(self):
+        """ like dict """
         return self.__dict__.__repr__()
     
     def __getitem__(self,key):
+        """ like dict """
         return self.__dict__.__getitem__(key)
-    
+
+    def __setattr__(self, key, value):
+        """ assigns a value, including functions which will becomes methods, e.g. the first argument must be 'self' """
+        self.__setitem__(key,value)
+
     def __setitem__(self,key,value):
+        """ assigns a value, including functions which will becomes methods, e.g. the first argument must be 'self' """
+        if isinstance(value,types.FunctionType):
+            # bind function to this object
+            value = types.MethodType(value,self)
+        elif isinstance(value,types.MethodType):
+            # re-point the method to the current instance
+            value = types.MethodType(value.__func__,self)
         self.__dict__.__setitem__(key,value)
         
+    def __len__(self):
+        return self.__dict__.__len__()
+        
     def __delitem__(self,key):
+        """ like dict """
         self.__dict__.__delitem__(key)
 
     def __iter__(self):
+        """ like dict """
         return self.__dict__.__iter__()
     
     def __contains__(self, key):
+        """ implements 'in' operator """
         return self.__dict__.__contains__(key)
     
     def __call__(self, key, *kargs):
+        """ Short-cut for get(), e.g. use __call__('a', 0) to ask for the value of 'a' with default 0 """
         return self.get(key,*kargs)
     
     def keys(self):
+        """ like dict """
         return self.__dict__.keys()
 
     def get(self, key, *kargs):
-        """ like dict.get() """
+        """ like dict.get(), e.g. get('a',0) gets the value of 'a' defaulting to 0. If no default is proivded, and 'a' is not defined the function fails """
         if len(kargs) == 0 or key in self:
             return self[key]
         if len(kargs) != 1:
@@ -330,30 +377,33 @@ class Generic(object):
         return kargs[0]
     
     def __add__(self, o):
-        """ Allows merging dicts or other Generics. To add other objects, use merge() explicitly """
-        out = Generic(self)
-        if isinstance(o,dict):
-            out.__dict__.update(o)
-        elif isinstance(o,Generic):
-            out.__dict__.update(o.__dict__)
-        else:
+        """ Allows merging dicts or other Generics into the Generic
+                g = Generic(a=1)
+                g += Generic(a=2,b=1)
+            or
+                g += {'a':2,'b':1)
+            gives a Generic witha=2,b=1.
+        """        
+        if not isinstance(o,(dict,Generic)):
             raise ValueError("Cannot handle type %s. Use merge() explicitly to add object members" % type(o))
-        return out
+        return Generic(self,o)
     
     def merge(self,*vargs,**kwargs):
         """ merges various data into the generic.
-                varargs: for each element, add the key/value pairs of dictionaries, Generics or any object (in the latter case __dict__ is used for non-functions)
+                varargs: for each element, add the key/value pairs of dictionaries, Generics or any object. For Generics, the function re-points methods
                 kwargs: are kwarg name/value p[airs all added directly, e.g. merge(a=1) adds 'a' with a value of '1' to self
                 
             Example
                 in1 = { 'a':1, 'b':2 }
                 in2 = Generic(c=3,d=4)
+                def F(self,x):
+                    self.d = x
+                in2.F = F   <-- creates a method
                 def O(object):
                     def __init__(self):
                         self.e = 5
-                in3 = O()
-                
-                merge( in1, in2, in3, f=6 )
+                in3 = O()                
+                merge( in1, in2, in3, g=6 )
                 
             Note
                 The function first processes the 'vargs' list, and then
@@ -364,13 +414,16 @@ class Generic(object):
             if isinstance(o,dict):
                 self.__dict__.update(o)
             elif isinstance(o,Generic):
-                self.__dict__.update(o.__dict__)
+                dct = getattr(o,"__dict__",None)
+                for e in dct:
+                    # this function handles functions and methods correctly.
+                    self.__setitem__(e,dct[e])        
             else:
                 dct = getattr(o,"__dict__",None)
                 if dct is None:
                     raise ValueError("Cannot handle type %s. It has no __dict__",type(o))
                 for e in dct:
-                    if not isFunction(e):
+                    if not isFunction(e):   # no automatic transfer of functions and methods from mere objects
                         self.__dict__.__setitem__(e,dct[e])        
         for o in vargs:
             merge_object_dict(o)
